"""Example datasets that are available for automatic downloading."""

AVAILABLE_CASES: dict[str, str] = {
    "example-input-volta": "10.4121/bbe10a2a-39dc-4098-a69f-0f677d06ecdd.v2",
    "example-input-eiffel": "10.4121/f9572240-f179-4338-9e1b-82c5598529e2.v1",
}
